<p class="sphinx-version">
    {% trans sphinx_version=sphinx_version|e %}Created using <a href="https://www.sphinx-doc.org/">Sphinx</a> {{ sphinx_version }}.{% endtrans %}
    <br/>
    </p>
    <p class="theme-version">
    {{ _("Built with the") }}
    <a href="https://pydata-sphinx-theme.readthedocs.io/en/stable/index.html">PyData Sphinx Theme</a>
    {{ theme_version }}.
    </p>
