<div class="things-in-a-row">
    <a href="https://neuroinformatics.dev/">
        <img src="{{ pathto('_static/light-logo-niu.png', 1) }}" alt="Sponsors" class="only-light img-sponsor"/>
        <img src="{{ pathto('_static/dark-logo-niu.png', 1) }}" alt="Sponsors" class="only-dark img-sponsor"/>
    </a>
    <a href="https://www.sainsburywellcome.org/web/">
        <img src="{{ pathto('_static/light-logo-swc.png', 1) }}" alt="Sponsors" class="only-light img-sponsor"/>
        <img src="{{ pathto('_static/dark-logo-swc.png', 1) }}" alt="Sponsors" class="only-dark img-sponsor"/>
    </a>
    <a href="https://www.ucl.ac.uk/gatsby/gatsby-computational-neuroscience-unit">
        <img src="{{ pathto('_static/light-logo-gatsby.png', 1) }}" alt="Sponsors" class="only-light img-sponsor"/>
        <img src="{{ pathto('_static/dark-logo-gatsby.png', 1) }}" alt="Sponsors" class="only-dark img-sponsor"/>
    </a>
    <a href="https://www.ucl.ac.uk/">
        <img src="{{ pathto('_static/light-logo-ucl.png', 1) }}" alt="Sponsors" class="only-light img-sponsor"/>
        <img src="{{ pathto('_static/dark-logo-ucl.png', 1) }}" alt="Sponsors" class="only-dark img-sponsor"/>
    </a>
</div>
