html[data-theme=dark] {
    --pst-color-primary: #04B46D;
    --pst-color-link: var(--pst-color-primary);
  }

html[data-theme=light] {
  --pst-color-primary: #03A062;
  --pst-color-link: var(--pst-color-primary);
}

body .bd-article-container {
max-width: 100em !important;
}

.col {
flex: 0 0 50%;
max-width: 50%;
}

.img-sponsor {
height: 50px;
padding-top: 5px;
padding-right: 5px;
padding-bottom: 5px;
padding-left: 5px;
}

.things-in-a-row {
display: flex;
flex-wrap: wrap;
justify-content: space-between;
}
